@@ -1,30 +1,29 @@
 /* auth.c
 ** strophe XMPP client library -- auth functions and handlers
 **
-** Copyright (C) 2005-2009 Collecta, Inc. 
+** Copyright (C) 2005-2009 Collecta, Inc.
 **
 **  This software is provided AS-IS with no warranty, either express or
 **  implied.
 **
-**  This software is distributed under license and may not be copied,
-**  modified or distributed except as expressly authorized under the
-**  terms of the license contained in the file LICENSE.txt in this
-**  distribution.
+** This program is dual licensed under the MIT and GPLv3 licenses.
 */
 
-/** @file 
+/** @file
  *  Authentication function and handlers.
  */
 
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <time.h>
 
 #include "strophe.h"
 #include "common.h"
 #include "sasl.h"
+#include "sha1.h"
 
-#ifdef _WIN32
+#ifdef _MSC_VER
 #define strcasecmp stricmp
 #endif
 
@@ -54,10 +53,21 @@
  */
 #define LEGACY_TIMEOUT 15000 /* 15 seconds */
 #endif
+#ifndef HANDSHAKE_TIMEOUT
+/** @def HANDSHAKE_TIMEOUT
+ *  Time to wait for component authentication to complete
+ */
+#define HANDSHAKE_TIMEOUT 15000 /* 15 seconds */
+#endif
 
 static void _auth(xmpp_conn_t * const conn);
-static void _handle_open_tls(xmpp_conn_t * const conn);
 static void _handle_open_sasl(xmpp_conn_t * const conn);
+
+static int _handle_component_auth(xmpp_conn_t * const conn);
+static int _handle_component_hs_response(xmpp_conn_t * const conn,
+            xmpp_stanza_t * const stanza,
+            void * const userdata);
+
 static int _handle_missing_legacy(xmpp_conn_t * const conn,
 				  void * const userdata);
 static int _handle_legacy(xmpp_conn_t * const conn,
@@ -75,6 +85,10 @@ static int _handle_digestmd5_challenge(xmpp_conn_t * const conn,
 static int _handle_digestmd5_rspauth(xmpp_conn_t * const conn,
 			xmpp_stanza_t * const stanza,
 			void * const userdata);
+static int _handle_scram_sha1_challenge(xmpp_conn_t * const conn,
+			xmpp_stanza_t * const stanza,
+			void * const userdata);
+static char *_make_scram_sha1_init_msg(xmpp_conn_t * const conn);
 
 static int _handle_missing_features_sasl(xmpp_conn_t * const conn,
 					 void * const userdata);
@@ -88,6 +102,8 @@ static int _handle_session(xmpp_conn_t * const conn,
 			   void * const userdata);
 static int _handle_missing_session(xmpp_conn_t * const conn,
 				   void * const userdata);
+static int _handle_missing_handshake(xmpp_conn_t * const conn,
+                                     void * const userdata);
 
 /* stream:error handler */
 static int _handle_error(xmpp_conn_t * const conn,
@@ -100,7 +116,7 @@ static int _handle_error(xmpp_conn_t * const conn,
     /* free old stream error if it's still there */
     if (conn->stream_error) {
 	xmpp_stanza_release(conn->stream_error->stanza);
-	if (conn->stream_error->text) 
+	if (conn->stream_error->text)
 	    xmpp_free(conn->ctx, conn->stream_error->text);
 	xmpp_free(conn->ctx, conn->stream_error);
     }
@@ -221,14 +237,16 @@ static int _handle_features(xmpp_conn_t * const conn,
     /* check for SASL */
     child = xmpp_stanza_get_child_by_name(stanza, "mechanisms");
     if (child && (strcmp(xmpp_stanza_get_ns(child), XMPP_NS_SASL) == 0)) {
-	for (mech = xmpp_stanza_get_children(child); mech; 
+	for (mech = xmpp_stanza_get_children(child); mech;
 	     mech = xmpp_stanza_get_next(mech)) {
-	    if (strcmp(xmpp_stanza_get_name(mech), "mechanism") == 0) {
+	    if (xmpp_stanza_get_name(mech) && strcmp(xmpp_stanza_get_name(mech), "mechanism") == 0) {
 		text = xmpp_stanza_get_text(mech);
 		if (strcasecmp(text, "PLAIN") == 0)
 		    conn->sasl_support |= SASL_MASK_PLAIN;
 		else if (strcasecmp(text, "DIGEST-MD5") == 0)
 		    conn->sasl_support |= SASL_MASK_DIGESTMD5;
+                else if (strcasecmp(text, "SCRAM-SHA-1") == 0)
+                    conn->sasl_support |= SASL_MASK_SCRAMSHA1;
 		else if (strcasecmp(text, "ANONYMOUS") == 0)
 		    conn->sasl_support |= SASL_MASK_ANONYMOUS;
 
@@ -238,7 +256,7 @@ static int _handle_features(xmpp_conn_t * const conn,
     }
 
     _auth(conn);
- 
+
     return 0;
 }
 
@@ -263,7 +281,7 @@ static int _handle_proceedtls_default(xmpp_conn_t * const conn,
 {
     char *name;
     name = xmpp_stanza_get_name(stanza);
-    xmpp_debug(conn->ctx, "xmpp", 
+    xmpp_debug(conn->ctx, "xmpp",
 	"handle proceedtls called for %s", name);
 
     if (strcmp(name, "proceed") == 0) {
@@ -277,7 +295,7 @@ static int _handle_proceedtls_default(xmpp_conn_t * const conn,
 	    tls_free(conn->tls);
 	    conn->tls = NULL;
 	    conn->tls_failed = 1;
-	
+
 	    /* failed tls spoils the connection, so disconnect */
 	    xmpp_disconnect(conn);
 	}
@@ -303,14 +321,14 @@ static int _handle_sasl_result(xmpp_conn_t * const conn,
 
     /* the server should send a <success> or <failure> stanza */
     if (strcmp(name, "failure") == 0) {
-	xmpp_debug(conn->ctx, "xmpp", "SASL %s auth failed", 
+	xmpp_debug(conn->ctx, "xmpp", "SASL %s auth failed",
 		   (char *)userdata);
-	
+
 	/* fall back to next auth method */
 	_auth(conn);
     } else if (strcmp(name, "success") == 0) {
 	/* SASL PLAIN auth successful, we need to restart the stream */
-	xmpp_debug(conn->ctx, "xmpp", "SASL %s auth successful", 
+	xmpp_debug(conn->ctx, "xmpp", "SASL %s auth successful",
 		   (char *)userdata);
 
 	/* reset parser */
@@ -355,10 +373,10 @@ static int _handle_digestmd5_challenge(xmpp_conn_t * const conn,
 	if (!auth) {
 	    disconnect_mem_error(conn);
 	    return 0;
-	}	
+	}
 	xmpp_stanza_set_name(auth, "response");
 	xmpp_stanza_set_ns(auth, XMPP_NS_SASL);
-	
+
 	authdata = xmpp_stanza_new(conn->ctx);
 	if (!authdata) {
 	    disconnect_mem_error(conn);
@@ -371,7 +389,7 @@ static int _handle_digestmd5_challenge(xmpp_conn_t * const conn,
 	xmpp_stanza_add_child(auth, authdata);
 	xmpp_stanza_release(authdata);
 
-	handler_add(conn, _handle_digestmd5_rspauth, 
+	handler_add(conn, _handle_digestmd5_rspauth,
 		    XMPP_NS_SASL, NULL, NULL, NULL);
 
 	xmpp_send(conn, auth);
@@ -404,7 +422,7 @@ static int _handle_digestmd5_rspauth(xmpp_conn_t * const conn,
 	if (!auth) {
 	    disconnect_mem_error(conn);
 	    return 0;
-	}	
+	}
 	xmpp_stanza_set_name(auth, "response");
 	xmpp_stanza_set_ns(auth, XMPP_NS_SASL);
 	xmpp_send(conn, auth);
@@ -416,6 +434,117 @@ static int _handle_digestmd5_rspauth(xmpp_conn_t * const conn,
     return 1;
 }
 
+/* handle the challenge phase of SCRAM-SHA-1 auth */
+static int _handle_scram_sha1_challenge(xmpp_conn_t * const conn,
+					xmpp_stanza_t * const stanza,
+					void * const userdata)
+{
+    char *text;
+    char *response;
+    xmpp_stanza_t *auth, *authdata;
+    char *name;
+    char *challenge;
+    char *scram_init = (char *)userdata;
+
+    name = xmpp_stanza_get_name(stanza);
+    xmpp_debug(conn->ctx, "xmpp",
+               "handle SCRAM-SHA-1 (challenge) called for %s", name);
+
+    if (strcmp(name, "challenge") == 0) {
+        text = xmpp_stanza_get_text(stanza);
+        if (!text)
+            goto err;
+
+        challenge = (char *)base64_decode(conn->ctx, text, strlen(text));
+        xmpp_free(conn->ctx, text);
+        if (!challenge)
+            goto err;
+
+        response = sasl_scram_sha1(conn->ctx, challenge, scram_init,
+                                   conn->jid, conn->pass);
+        xmpp_free(conn->ctx, challenge);
+        if (!response)
+            goto err;
+
+        auth = xmpp_stanza_new(conn->ctx);
+        if (!auth)
+            goto err_free_response;
+        xmpp_stanza_set_name(auth, "response");
+        xmpp_stanza_set_ns(auth, XMPP_NS_SASL);
+
+        authdata = xmpp_stanza_new(conn->ctx);
+        if (!authdata)
+            goto err_release_auth;
+        xmpp_stanza_set_text(authdata, response);
+        xmpp_free(conn->ctx, response);
+
+        xmpp_stanza_add_child(auth, authdata);
+        xmpp_stanza_release(authdata);
+
+        xmpp_send(conn, auth);
+        xmpp_stanza_release(auth);
+
+    } else {
+        xmpp_free(conn->ctx, scram_init);
+        return _handle_sasl_result(conn, stanza, "SCRAM-SHA-1");
+    }
+
+    return 1;
+
+err_release_auth:
+    xmpp_stanza_release(auth);
+err_free_response:
+    xmpp_free(conn->ctx, response);
+err:
+    xmpp_free(conn->ctx, scram_init);
+    disconnect_mem_error(conn);
+    return 0;
+}
+
+static char *_get_nonce(xmpp_ctx_t *ctx)
+{
+    unsigned char buffer[sizeof(clock_t) + sizeof(time_t)] = {0};
+    clock_t ticks = clock();
+    time_t t;
+
+    if (ticks != (clock_t)-1) {
+        *(clock_t *)buffer = ticks;
+    }
+    t = time((time_t *)(buffer + sizeof(clock_t)));
+    if (t == (time_t)-1) {
+        *(time_t *)(buffer + sizeof(clock_t)) = (time_t)rand();
+    }
+
+    return base64_encode(ctx, buffer, sizeof(buffer));
+}
+
+static char *_make_scram_sha1_init_msg(xmpp_conn_t * const conn)
+{
+    size_t message_len;
+    char *node;
+    char *message;
+    char *nonce;
+
+    node = xmpp_jid_node(conn->ctx, conn->jid);
+    if (!node) {
+        return NULL;
+    }
+
+    nonce = _get_nonce(conn->ctx);
+    if (!nonce) {
+        return NULL;
+    }
+    message_len = strlen(node) + strlen(nonce) + 8 + 1;
+    message = xmpp_alloc(conn->ctx, message_len);
+    if (message) {
+        xmpp_snprintf(message, message_len, "n,,n=%s,r=%s", node, nonce);
+        xmpp_free(conn->ctx, node);
+    }
+    xmpp_free(conn->ctx, nonce);
+
+    return message;
+}
+
 static xmpp_stanza_t *_make_starttls(xmpp_conn_t * const conn)
 {
     xmpp_stanza_t *starttls;
@@ -426,7 +555,7 @@ static xmpp_stanza_t *_make_starttls(xmpp_conn_t * const conn)
 	xmpp_stanza_set_name(starttls, "starttls");
 	xmpp_stanza_set_ns(starttls, XMPP_NS_TLS);
     }
-    
+
     return starttls;
 }
 
@@ -442,19 +571,20 @@ static xmpp_stanza_t *_make_sasl_auth(xmpp_conn_t * const conn,
 	xmpp_stanza_set_ns(auth, XMPP_NS_SASL);
 	xmpp_stanza_set_attribute(auth, "mechanism", mechanism);
     }
-    
+
     return auth;
 }
 
-/* authenticate the connection 
- * this may get called multiple times.  if any auth method fails, 
+/* authenticate the connection
+ * this may get called multiple times.  if any auth method fails,
  * this will get called again until one auth method succeeds or every
- * method fails 
+ * method fails
  */
 static void _auth(xmpp_conn_t * const conn)
 {
     xmpp_stanza_t *auth, *authdata, *query, *child, *iq;
     char *str, *authid;
+    char *scram_init;
     int anonjid;
 
     /* if there is no node in conn->jid, we assume anonymous connect */
@@ -489,7 +619,7 @@ static void _auth(xmpp_conn_t * const conn)
 	    return;
 	}
 
-	handler_add(conn, _handle_proceedtls_default, 
+	handler_add(conn, _handle_proceedtls_default,
 		    XMPP_NS_TLS, NULL, NULL, NULL);
 
 	xmpp_send(conn, auth);
@@ -514,9 +644,54 @@ static void _auth(xmpp_conn_t * const conn)
 	/* SASL ANONYMOUS was tried, unset flag */
 	conn->sasl_support &= ~SASL_MASK_ANONYMOUS;
     } else if (anonjid) {
-	xmpp_error(conn->ctx, "auth", 
+	xmpp_error(conn->ctx, "auth",
 		   "No node in JID, and SASL ANONYMOUS unsupported.");
 	xmpp_disconnect(conn);
+    } else if (conn->sasl_support & SASL_MASK_SCRAMSHA1) {
+        auth = _make_sasl_auth(conn, "SCRAM-SHA-1");
+        if (!auth) {
+            disconnect_mem_error(conn);
+            return;
+        }
+
+        /* don't free scram_init on success */
+        scram_init = _make_scram_sha1_init_msg(conn);
+        if (!scram_init) {
+            xmpp_stanza_release(auth);
+            disconnect_mem_error(conn);
+            return;
+        }
+
+        str = (char *)base64_encode(conn->ctx, (unsigned char *)scram_init,
+                                    strlen(scram_init));
+        if (!str) {
+            xmpp_free(conn->ctx, scram_init);
+            xmpp_stanza_release(auth);
+            disconnect_mem_error(conn);
+            return;
+        }
+
+        authdata = xmpp_stanza_new(conn->ctx);
+        if (!authdata) {
+            xmpp_free(conn->ctx, str);
+            xmpp_free(conn->ctx, scram_init);
+            xmpp_stanza_release(auth);
+            disconnect_mem_error(conn);
+            return;
+        }
+        xmpp_stanza_set_text(authdata, str);
+        xmpp_free(conn->ctx, str);
+        xmpp_stanza_add_child(auth, authdata);
+        xmpp_stanza_release(authdata);
+
+        handler_add(conn, _handle_scram_sha1_challenge,
+                    XMPP_NS_SASL, NULL, NULL, (void *)scram_init);
+
+        xmpp_send(conn, auth);
+        xmpp_stanza_release(auth);
+
+        /* SASL SCRAM-SHA-1 was tried, unset flag */
+        conn->sasl_support &= ~SASL_MASK_SCRAMSHA1;
     } else if (conn->sasl_support & SASL_MASK_DIGESTMD5) {
 	auth = _make_sasl_auth(conn, "DIGEST-MD5");
 	if (!auth) {
@@ -525,7 +700,7 @@ static void _auth(xmpp_conn_t * const conn)
 
 	}
 
-	handler_add(conn, _handle_digestmd5_challenge, 
+	handler_add(conn, _handle_digestmd5_challenge,
 		    XMPP_NS_SASL, NULL, NULL, NULL);
 
 	xmpp_send(conn, auth);
@@ -543,7 +718,7 @@ static void _auth(xmpp_conn_t * const conn)
 	if (!authdata) {
 	    disconnect_mem_error(conn);
 	    return;
-	}	
+	}
 	authid = _get_authid(conn);
 	if (!authid) {
 	    disconnect_mem_error(conn);
@@ -556,6 +731,7 @@ static void _auth(xmpp_conn_t * const conn)
 	}
 	xmpp_stanza_set_text(authdata, str);
 	xmpp_free(conn->ctx, str);
+	xmpp_free(conn->ctx, authid);
 
 	xmpp_stanza_add_child(auth, authdata);
 	xmpp_stanza_release(authdata);
@@ -570,7 +746,7 @@ static void _auth(xmpp_conn_t * const conn)
 	conn->sasl_support &= ~SASL_MASK_PLAIN;
     } else if (conn->type == XMPP_CLIENT) {
 	/* legacy client authentication */
-	
+
 	iq = xmpp_stanza_new(conn->ctx);
 	if (!iq) {
 	    disconnect_mem_error(conn);
@@ -656,7 +832,7 @@ static void _auth(xmpp_conn_t * const conn)
 	} else {
 	    xmpp_stanza_release(authdata);
 	    xmpp_stanza_release(iq);
-	    xmpp_error(conn->ctx, "auth", 
+	    xmpp_error(conn->ctx, "auth",
 		       "Cannot authenticate without resource");
 	    xmpp_disconnect(conn);
 	    return;
@@ -665,7 +841,7 @@ static void _auth(xmpp_conn_t * const conn)
 	xmpp_stanza_release(authdata);
 
 	handler_add_id(conn, _handle_legacy, "_xmpp_auth1", NULL);
-	handler_add_timed(conn, _handle_missing_legacy, 
+	handler_add_timed(conn, _handle_missing_legacy,
 			  LEGACY_TIMEOUT, NULL);
 
 	xmpp_send(conn, iq);
@@ -677,7 +853,7 @@ static void _auth(xmpp_conn_t * const conn)
 /** Set up handlers at stream start.
  *  This function is called internally to Strophe for handling the opening
  *  of an XMPP stream.  It's called by the parser when a stream is opened
- *  or reset, and adds the initial handlers for <stream:error/> and 
+ *  or reset, and adds the initial handlers for <stream:error/> and
  *  <stream:features/>.  This function is not intended for use outside
  *  of Strophe.
  *
@@ -690,11 +866,11 @@ void auth_handle_open(xmpp_conn_t * const conn)
 
     /* setup handler for stream:error */
     handler_add(conn, _handle_error,
-		NULL, "stream:error", NULL, NULL);
+		XMPP_NS_STREAMS, "error", NULL, NULL);
 
     /* setup handlers for incoming <stream:features> */
     handler_add(conn, _handle_features,
-		NULL, "stream:features", NULL, NULL);
+		XMPP_NS_STREAMS, "features", NULL, NULL);
     handler_add_timed(conn, _handle_missing_features,
 		      FEATURES_TIMEOUT, NULL);
 }
@@ -706,7 +882,7 @@ static void _handle_open_sasl(xmpp_conn_t * const conn)
 
     /* setup stream:features handlers */
     handler_add(conn, _handle_features_sasl,
-		NULL, "stream:features", NULL, NULL);
+		XMPP_NS_STREAMS, "features", NULL, NULL);
     handler_add_timed(conn, _handle_missing_features_sasl,
 		      FEATURES_TIMEOUT, NULL);
 }
@@ -739,7 +915,7 @@ static int _handle_features_sasl(xmpp_conn_t * const conn,
     /* if bind is required, go ahead and start it */
     if (conn->bind_required) {
 	/* bind resource */
-	
+
 	/* setup response handlers */
 	handler_add_id(conn, _handle_bind, "_xmpp_bind1", NULL);
 	handler_add_timed(conn, _handle_missing_bind,
@@ -771,7 +947,7 @@ static int _handle_features_sasl(xmpp_conn_t * const conn,
 	    resource = NULL;
 	}
 
-	/* if we have a resource to request, do it. otherwise the 
+	/* if we have a resource to request, do it. otherwise the
 	   server will assign us one */
 	if (resource) {
 	    res = xmpp_stanza_new(conn->ctx);
@@ -822,7 +998,7 @@ static int _handle_missing_features_sasl(xmpp_conn_t * const conn,
     xmpp_disconnect(conn);
     return 0;
 }
-					  
+
 static int _handle_bind(xmpp_conn_t * const conn,
 			xmpp_stanza_t * const stanza,
 			void * const userdata)
@@ -854,7 +1030,7 @@ static int _handle_bind(xmpp_conn_t * const conn,
 	if (conn->session_required) {
 	    /* setup response handlers */
 	    handler_add_id(conn, _handle_session, "_xmpp_session1", NULL);
-	    handler_add_timed(conn, _handle_missing_session, 
+	    handler_add_timed(conn, _handle_missing_session,
 			      SESSION_TIMEOUT, NULL);
 
 	    /* send session request */
@@ -885,9 +1061,9 @@ static int _handle_bind(xmpp_conn_t * const conn,
 	    xmpp_stanza_release(iq);
 	} else {
 	    conn->authenticated = 1;
-	   
+
 	    /* call connection handler */
-	    conn->conn_handler(conn, XMPP_CONN_CONNECT, 0, NULL, 
+	    conn->conn_handler(conn, XMPP_CONN_CONNECT, 0, NULL,
 			       conn->userdata);
 	}
     } else {
@@ -924,7 +1100,7 @@ static int _handle_session(xmpp_conn_t * const conn,
 	xmpp_debug(conn->ctx, "xmpp", "Session establishment successful.");
 
 	conn->authenticated = 1;
-	
+
 	/* call connection handler */
 	conn->conn_handler(conn, XMPP_CONN_CONNECT, 0, NULL, conn->userdata);
     } else {
@@ -987,3 +1163,96 @@ static int _handle_missing_legacy(xmpp_conn_t * const conn,
     return 0;
 }
 
+void auth_handle_component_open(xmpp_conn_t * const conn)
+{
+    /* reset all timed handlers */
+    handler_reset_timed(conn, 0);
+
+    handler_add(conn, _handle_error, XMPP_NS_STREAMS, "error", NULL, NULL);
+    handler_add(conn, _handle_component_hs_response, NULL,
+                "handshake", NULL, NULL);
+    handler_add_timed(conn, _handle_missing_handshake, HANDSHAKE_TIMEOUT, NULL);
+
+    _handle_component_auth(conn);
+}
+
+/* Will compute SHA1 and authenticate the component to the server */
+int _handle_component_auth(xmpp_conn_t * const conn)
+{
+    uint8_t md_value[SHA1_DIGEST_SIZE];
+    SHA1_CTX mdctx;
+    char *digest;
+    size_t i;
+
+    /* Feed the session id and passphrase to the algorithm.
+     * We need to compute SHA1(session_id + passphrase)
+     */
+    SHA1_Init(&mdctx);
+    SHA1_Update(&mdctx, (uint8_t*)conn->stream_id, strlen(conn->stream_id));
+    SHA1_Update(&mdctx, (uint8_t*)conn->pass, strlen(conn->pass));
+    SHA1_Final(&mdctx, md_value);
+
+    digest = xmpp_alloc(conn->ctx, 2*sizeof(md_value)+1);
+    if (digest) {
+        /* convert the digest into string representation */
+        for (i = 0; i < sizeof(md_value); i++)
+            xmpp_snprintf(digest+i*2, 3, "%02x", md_value[i]);
+        digest[2*sizeof(md_value)] = '\0';
+
+        xmpp_debug(conn->ctx, "auth", "Digest: %s, len: %d",
+                   digest, strlen(digest));
+
+        /* Send the digest to the server */
+        xmpp_send_raw_string(conn, "<handshake xmlns='%s'>%s</handshake>",
+                             XMPP_NS_COMPONENT, digest);
+        xmpp_debug(conn->ctx, "auth", "Sent component handshake to the server.");
+        xmpp_free(conn->ctx, digest);
+    } else {
+        xmpp_debug(conn->ctx, "auth", "Couldn't allocate memory for component "\
+                                      "handshake digest.");
+        xmpp_disconnect(conn);
+        return XMPP_EMEM;
+    }
+
+    return 0;
+}
+
+/* Check if the received stanza is <handshake/> and set auth to true
+ * and fire connection handler.
+ */
+int _handle_component_hs_response(xmpp_conn_t * const conn,
+            xmpp_stanza_t * const stanza,
+            void * const userdata)
+{
+    char *name;
+
+    xmpp_timed_handler_delete(conn, _handle_missing_handshake);
+
+    name = xmpp_stanza_get_name(stanza);
+    if (strcmp(name, "handshake") != 0) {
+        char *msg;
+        size_t msg_size;
+        xmpp_stanza_to_text(stanza, &msg, &msg_size);
+        if (msg) {
+            xmpp_debug(conn->ctx, "auth", "Handshake failed: %s", msg);
+            xmpp_free(conn->ctx, msg);
+        }
+        xmpp_disconnect(conn);
+        return XMPP_EINT;
+    } else {
+        conn->authenticated = 1;
+        conn->conn_handler(conn, XMPP_CONN_CONNECT, 0, NULL, conn->userdata);
+    }
+
+    /* We don't need this handler anymore, return 0 so it can be deleted
+     * from the list of handlers.
+     */
+    return 0;
+}
+
+int _handle_missing_handshake(xmpp_conn_t * const conn, void * const userdata)
+{
+    xmpp_error(conn->ctx, "xmpp", "Server did not reply to handshake request.");
+    xmpp_disconnect(conn);
+    return 0;
+}
